import uuid, os, xmlrpclib, subprocess, logging
import pkg_resources
from string import Template
from optparse import OptionParser
import sys
import socket
import MySQLdb
import time

SUPERVISORD = os.path.join('var', 'bin', 'supervisord')
SUPERVISORCTL = os.path.join('var', 'bin', 'supervisorctl')

class XMLRPCServer(object):
  def __init__(self, server_url, client_key):
    self.client_key = client_key
    self.server_url = server_url

  def call(self, method, *args):
    server = xmlrpclib.ServerProxy(self.server_url)
    args = list(args)
    args.insert(0, self.client_key)
    return getattr(server, method)(*args)

class PercentTemplate(Template):
  delimiter = '%'

def generateServerKey():
  return str(uuid.uuid4())

def getServerKey(key_file):
  # TODO:
  #  * make key safe (600)
  #  * make generation safe (remove file in case of problem)
  if os.path.exists(key_file):
    return file(key_file).readlines()[0].strip()
  new_uid = generateServerKey()
  file(key_file, 'w').write(new_uid)
  return new_uid

class Partition(object):
  # generic
  def notImplemented(self):
    """Simple not implemented wrapper"""
    return None

  def __init__(self, partition_dict):
    self.item_id = partition_dict['ID']
    self.partition_id = partition_dict['PARTITION_ID']
    self.partition_type = partition_dict['TYPE'].replace(' ', '')
    self.partition_dict = partition_dict
    plugin_init = getattr(self, '%s' % self.partition_type, None)
    if plugin_init is None:
      # plugin not supported, create dummy method doing nothing
      logging.warning('Partition management plugin for %s not found, '
        'generating dummy one' % self.partition_type)
      for method in 'install', 'start', 'stop', 'getPartitionState':
        setattr(self, '%s_%s' % (method, self.partition_type),
            self.notImplemented)
    else:
      plugin_init()

  def install(self):
    return getattr(self, 'install_%s' % self.partition_type)()

  def start(self):
    return getattr(self, 'start_%s' % self.partition_type)()

  def stop(self):
    return getattr(self, 'stop_%s' % self.partition_type)()

  def getPartitionState(self):
    return getattr(self, 'getPartitionState_%s' % self.partition_type)()

  # Specific code - candidates for plugins
  # Business Configuration
  def BusinessConfiguration(self):
    pass

  def install_BusinessConfiguration(self):
    return self.getPartitionState_BusinessConfiguration()

  def start_BusinessConfiguration(self):
    return self.getPartitionState_BusinessConfiguration()

  def stop_BusinessConfiguration(self):
    return self.getPartitionState_BusinessConfiguration()

  def connectBusinessConfiguration(self):
    self.connection = xmlrpclib.ServerProxy(
        self.connection_string,
        allow_none=True)

  def getPartitionState_BusinessConfiguration(self):
    self.connection_string = '%(PROTOCOL)s://%(USERNAME)s:%(PASSWORD)s@%(HOS'\
      'TNAME)s:%(TCP_PORT)s/%(RESOURCE_ID)s/' % \
      self.partition_dict['CONNECTION_DICT']
    self.connectBusinessConfiguration()
    try:
      # if site was configured, we shall not do nothing
      # XXX-Luke: This is against "work always" assumption
      try:
        self.connection.portal_preferences.default_configurator_preference \
            .getId()
      except xmlrpclib.Fault, e:
        # not configured
        if 'zExceptions.NotFound' in e.faultString:
          pass
      else:
        # configured, do not touch
        logging.info('%s already configured.' % self.partition_dict['ID'])
        return 'reportStarted', 'Already configured'
      if self.connection.portal_activities.countMessageWithTag(
          'initialERP5Setup') > 0:
        # do not work on instance, on witch initial setup is done
        # this is one big transaction and timmy makes it restart
        logging.info('%s configuration in progress.' % self.partition_dict['ID'])
        return 'reportStarting', 'Configuration in progress'
      # set id
      self.connection.setERP5SiteGlobalId(self.partition_dict[
        'ID'])
      # install templates
      self.connection.portal_templates.updateRepositoryBusinessTemplateList(
          self.partition_dict['BT5_REPOSITORY_LIST'], None)
      for bt_id in self.partition_dict['BT5_LIST']:
        result = self.connection.portal_templates\
          .installBusinessTemplatesFromRepositories([bt_id])
        logging.info('%s: Installed Business Templates with result: %s' % (
        self.partition_dict['ID'], ','.join(result)))
      # XXX-Luke: In case if BT is installed, it might be required to restart

      # add special user
      try:
        self.connection.acl_users.zodb_users.manage_addUser(
          self.partition_dict['CONFIGURATOR_LOGIN'],'',
          self.partition_dict['CONFIGURATOR_PASSWORD'],
          self.partition_dict['CONFIGURATOR_PASSWORD'],
          None)
      except xmlrpclib.Fault, e:
        if 'Duplicate user ID' not in e.faultString:
          logging.error('%s: xmlrpclib.Fault: %s' % (self.partition_dict['ID'],
            e.faultString))
          return 'reportError', 'xmlrpclib.Fault: %s' % e.faultString
      # missing possiblity to do:
      #  self.connection.zodb_roles.doAssignRoleToPrincipal(
      #    self.partition_dict['CONFIGURATOR_LOGIN'], 'Member')
      self.connection.portal_wizard.manage_addLocalRoles(
          self.partition_dict['CONFIGURATOR_LOGIN'],
          ('Manager',), None)
      # preferences
      default_site_preference = self.connection.portal_preferences\
          .default_site_preference
      default_site_preference.setPreferredWitchToolServerUrl(
          self.partition_dict['WITCH_SERVER_URL'])
      default_site_preference.setPreferredWitchToolServerRoot(
          self.partition_dict['WITCH_SERVER_ROOT'])
      default_site_preference.setPreferredExpressSubscriptionStatus('1')
      default_site_preference.setPreferredExpressErp5Uid(
          self.partition_dict['ID'])
      default_site_preference.setPreferredHtmlStyleDevelopperMode(None)
      default_site_preference.setPreferredHtmlStyleAccessTab('1')
      if default_site_preference.getPreferenceState() == 'disabled':
        default_site_preference.enable()

      customer_preference = self.connection.portal_preferences\
          .erp5_express_default_customer_preference
      customer_preference.setPreferredExpressUserId(
        self.partition_dict['WITCH_LOGIN'])
      customer_preference.setPreferredExpressPassword(
        self.partition_dict['WITCH_PASSWORD'])
      customer_preference.manage_delLocalRoles(
          [self.partition_dict['INSTANCE_ADMIN_USER']], None)
      customer_preference.manage_addLocalRoles(self.partition_dict[
        'CONFIGURATOR_LOGIN'], ('Owner',), None)
      if customer_preference.getPreferenceState() == 'disabled':
        customer_preference.enable()
      logging.info('%s: prepared for configuration' % self.partition_dict['ID'])
      return 'reportInstalled', 'Ready for configuration'
    except socket.error:
      logging.warning('Cannot connect to %s, ignoring for now' %
          self.connection_string)
      return 'reportUnknown', 'Cannot connect to Business Configuration, ign'\
          'ored for now'
    except Exception, e:
      logging.error('Uncaught error %s for %s' % (e, self.connection_string))
      return 'reportError', 'Uncaught error %s' % e

  # Computer
  def Computer(self):
    pass

  def install_Computer(self):
    pass

  def start_Computer(self):
    pass

  def stop_Computer(self):
    pass

  def getPartitionState_Computer(self):
    pass

  # Mysql Database
  def MysqlDatabase(self):
    pass

  def install_MysqlDatabase(self):
    return self.getPartitionState_MysqlDatabase()

  def start_MysqlDatabase(self):
    return self.getPartitionState_MysqlDatabase()

  def stop_MysqlDatabase(self):
    return self.getPartitionState_MysqlDatabase()

  def getPartitionState_MysqlDatabase(self):
    connection = self.partition_dict['CONNECTION_DICT']
    try:
      MySQLdb.connect(host=connection['HOSTNAME'] or connection['IP_ADDRESS'],
          port = int(connection['TCP_PORT']), user = connection['USERNAME'],
          passwd=connection['PASSWORD'])
    except MySQLdb.OperationalError, e:
      return 'reportError', 'Problem with connection %s' % e
    else:
      return 'reportStarted', 'Database accessible'

  # Mysql Server
  def MysqlServer(self):
    self.control_script = SUPERVISORCTL

  def install_MysqlServer(self):
    return self.getPartitionState_MysqlServer()

  def start_MysqlServer(self):
    return self.getPartitionState_MysqlServer()

  def stop_MysqlServer(self):
    return self.getPartitionState_MysqlServer()

  def getPartitionState_MysqlServer(self):
    return self.getPartitionState_Supervised('mysql')

  # Conversion Server
  def ConversionServer(self):
    self.control_script = SUPERVISORCTL

  def install_ConversionServer(self):
    return self.getPartitionState_ConversionServer()

  def start_ConversionServer(self):
    return self.getPartitionState_ConversionServer()

  def stop_ConversionServer(self):
    return self.getPartitionState_ConversionServer()

  def getPartitionState_Supervised(self, name):
    # XXX: Use XML RPC
    popen = subprocess.Popen([self.control_script, 'status', name],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (result_std, result_err) = popen.communicate()
    if 'RUNNING' in result_std:
      return 'reportStarted', 'supervised'
    if 'STARTING' in result_std:
      return 'reportStarting', 'supervised'
    if 'STOPPING' in result_std:
      return 'reportStopping', 'supervised'
    if 'STOPPED' in result_std:
      return 'reportStopped', 'supervised'
    return 'reportError', 'supervised: [std: %s] [err: %s]' % (result_std,
        result_err)

  def getPartitionState_ConversionServer(self):
    return self.getPartitionState_Supervised('oood')

  # Zope Instance related
  def ZopeInstance(self):
    self.control_script = os.path.join('var', self.item_id, 'bin', 'zopectl')

  def _zopectl_subprocess(self, action):
    if not os.path.exists(self.control_script):
      return ['reportError', 'no control script', '1']
    popen = subprocess.Popen([self.control_script, action],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return list(popen.communicate()) + [popen.returncode]

  def install_ZopeInstance(self):
    # XXX: Very simple way to know if partition was installed
    if self.getPartitionState_ZopeInstance() == 'started':
      self.stop_ZopeInstance()
      return 'reportError', 'Installation requested, instance was running - '\
          'instance stopped, on next run will be installed'
    if os.path.exists(self.control_script):
      return 'reportInstalled', ''
    else:
      return 'reportError', 'Installation failed'

  def start_ZopeInstance(self):
    if self.getPartitionState()[0] == 'started':
      return 'reportStarted', ''
    logging.info('Starting instance %s' % self.item_id)
    (result_std, result_err, returncode) = self._zopectl_subprocess('start')
    if returncode:
      return 'reportError', 'failed to start'
    return 'reportStarting', ''

  def stop_ZopeInstance(self):
    if self.getPartitionState()[0] == 'stopped':
      return 'reportStopped', ''
    logging.info('Stopping instance %s' % self.item_id)
    (result_std, result_err, returncode) = self._zopectl_subprocess('stop')
    if returncode:
      return 'reportError', 'failed to stop'
    return 'reportStopping', ''

  def getPartitionState_ZopeInstance(self):
    logging.info('Fetching status of instance %s' % self.item_id)
    (result_std, result_err, returncode) = self._zopectl_subprocess('status')
    if result_std.startswith('program running'):
      return 'started', ''
    elif 'not running' in result_std:
      return 'stopped', ''
    else:
      return 'error', '%s %s' % (result_std, result_err)

  # KVM Instance related
  def KVMInstance(self):
    self.control_script = os.path.join('bin', '%s-ctl' % self.item_id)

  install_KVMInstance = install_ZopeInstance
  start_KVMInstance = start_ZopeInstance
  stop_KVMInstance = stop_ZopeInstance

  def getPartitionState_KVMInstance(self):
    logging.info('Fetching status of instance %s' % self.item_id)
    (result_std, result_err, returncode) = self._zopectl_subprocess('status')
    if "True" in result_std:
      return 'started', ''
    elif 'False' in result_std:
      return 'stopped', ''
    else:
      return 'error', '%s %s' % (result_std, result_err)

  # Test Instance related
  def TestInstance(self):
    self.control_script = os.path.join('bin', '%s_run_test' % self.item_id)

  install_TestInstance = install_ZopeInstance
  start_TestInstance = start_ZopeInstance
  stop_TestInstance = stop_ZopeInstance

  getPartitionState_TestInstance = getPartitionState_KVMInstance

class Helper(object):
  def manageNothing(self, partition, server):
    pass

  def manageInstall(self, partition, server):
    result = partition.install()
    if result:
      server.call('updatePartitionState', partition.partition_id,
        *result)

  def manageStart(self, partition, server):
    result = partition.start()
    if result:
      server.call('updatePartitionState', partition.partition_id,
        *result)

  def manageStop(self, partition, server):
    result = partition.stop()
    if result:
      server.call('updatePartitionState', partition.partition_id,
        *result)

  def manageStatus(self, partition, server):
    result = partition.getPartitionState()
    if result:
      server.call('updatePartitionState', partition.partition_id,
        *result)

def keygenParseOptions():
  parser = OptionParser()
  parser.add_option("-k", "--key-file",
    help="File with server key")
  (options, args) = parser.parse_args()
  required_option_list = [
      'key_file',
    ]
  for o in required_option_list:
    if not getattr(options, o, None):
      parser.error('All options are required, missing "%s".' % o)
  return options, args

def keygen():
  (options, args) = keygenParseOptions()
  key = getServerKey(options.key_file)
  print 'Server key located at: %s' % os.path.abspath(options.key_file)
  print 'Key: %s' % key

class Timmy(object):
  def __init__(self):
    socket.setdefaulttimeout(30.0)
    self.options = self.parseOptions()[0]
    self.buildout = self.options.buildout_binary
    self.buildout_profile = self.options.main_output
    self.buildout_offline = self.options.buildout_offline
    logging_kw = dict(level=logging.INFO,
        format='%(asctime)s:%(levelname)s:%(name)s:%(message)s')
    if self.options.log_file is not None:
      logging_kw['filename'] = self.options.log_file
    logging.basicConfig(**logging_kw)
    self.time_begin = time.time()
    logging.info('[%s] Timmy started' % os.getpid())
    self.initialiseTemplateDirectoryList()

  def initialiseTemplateDirectoryList(self):
    """Returns list of directories, in which there might be templates"""
    self.template_directory_list = []
    if pkg_resources.resource_isdir(__name__, 'template'):
      template_directory = pkg_resources.resource_filename(__name__, "template")
      self.template_directory_list.append(template_directory)
      logging.info('Using %r as fallback directory with templates' %
          template_directory)

    if self.options.template_directory is not None:
      if os.path.isdir(self.options.template_directory):
        template_directory = os.path.abspath(self.options.template_directory)
        self.template_directory_list.append(template_directory)
        logging.info('Using %r as default directory with templates' %
            template_directory)
      else:
        logging.warning('User template %r is not a directory' %
            template_directory)
    if len(self.template_directory_list) == 0:
      logging.warning('No templates were found, some functionality will be not available')
    self.template_directory_list.reverse()
    return self.template_directory_list

  def run(self):
    try:
      self.setRunning()
    except:
      # XXX-Luke: If possible it would be nice to send such issue to
      #           master server
      logging.error('Issue during setting pid file %s %s' % sys.exc_info()[:2])
      raise
    try:
      try:
        try:
          self.server = XMLRPCServer(self.options.server_url,
              getServerKey(self.options.key_file))
          self.partition_dict_list = self.server.call('getComputerConfigurationDictList')
        except socket.error, e:
          logging.error('Unable to connect to remote server, verify if the URL is'
              ' accessible, failed: %s.' % e)
          sys.exit(1)
        except xmlrpclib.Fault, e:
          logging.error('Error found on server side, unable to continue, failed:'
              ' %s.' % e.faultString)
          sys.exit(1)

        self.computer_id = getServerKey(self.options.key_file)
        self.server.call('updatePartitionState', self.computer_id, 'reportStarted',
            'timmy is running on computer')
        self.supervisor_id = self.computer_id
        supervisor_list = [q for q in self.partition_dict_list if q['TYPE'] == \
            'Supervisor Server']
        if len(supervisor_list) > 0:
          self.supervisor_id = supervisor_list[0]['PARTITION_ID']

        # prepare - run supervisor
        try:
          supervisord_popen = subprocess.Popen([SUPERVISORD],
              stdout=subprocess.PIPE, stderr=subprocess.PIPE)
          (result_std, result_err) = supervisord_popen.communicate()
        except:
          self.server.call('updatePartitionState', self.supervisor_id, 'reportError',
              'Unsupported issue while trying to start supervisord: %s:%s' %
                sys.exc_info()[:2] )
          raise

        if supervisord_popen.returncode == 0:
          log_message = 'Supervisord started with: stdout = %r stderr = %r' % (
              result_std, result_err)
          self.server.call('updatePartitionState', self.supervisor_id, 'reportStarted',
              log_message)
          logging.info(log_message)
        else:
          if "Another program is already listening" in result_err:
            logging.info('Supervisord already running: stdout = %r stderr = '
                '%r' % (result_std, result_err))
          else:
            log_message = 'Supervisord unknown problem: stdout = %r stderr = %r' \
                % (result_std, result_err)
            self.server.call('updatePartitionState', self.supervisor_id, 'reportError',
                log_message)
            logging.info(log_message)

        # 1a pass - instance profiles
        self.updateInstanceProfiles()
        # 1b pass - main profile
        self.updateBaseProfile()
        # 2 pass - run buildout
        self.runBuildout()
        self.updateSupervisor()
        self.startAllSupervisored()

        # 3 pass - manage instances
        # XXX-Luke: To be moved to proper recipe
        updated_partition_list = []
        helper = Helper()
        for partition_dict in self.partition_dict_list:
          # install
          # stop
          # switch to special management
          # start
          partition = Partition(partition_dict)
          switcher_dict = dict(
              start = 'manageStart',
              stop = 'manageStop',
              nothing = 'manageNothing',
              install = 'manageInstall',
          )
          action = getattr(helper, switcher_dict[partition_dict['ACTION']])
          action(partition, self.server)
      except xmlrpclib.Fault, e:
        logging.error('Unexpected error in xml-rpc communication, unable to co'
            'ntinue, failed: %s.' % e.faultString)
        raise
      except socket.sslerror, e:
        logging.error('Unexpected error in socket communication, unable to co'
            'ntinue, failed: %s.' % e)
        raise
    finally:
      self.setFinished()
      logging.info('[%s] Timmy finished, invocation time: %.3fs' % (os.getpid(),
        time.time() - self.time_begin))

  def runBuildout(self):
    try:
      invoke_list = [self.buildout]
      if self.buildout_offline:
        invoke_list.append('-N')
      invoke_list.extend(['-c', self.buildout_profile, '-t', '5'])
      logging.info('invoking %s' % ' '.join(invoke_list))
      popen = subprocess.Popen([self.buildout, '-c', self.buildout_profile],
          stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      (result_std, result_err) = popen.communicate()
      # TODO: check that site is accesible in Data.fs, we cannot belive in
      #       Popen.returncode, as subprocesses of subprocesses are not handled
      #       properly

      # TODO: parse result_std and result_err

      if popen.returncode == 0:
        log_method = logging.info
      else:
        log_method = logging.error
        log_method('Buildout finished with bad status code (%s)' % \
            popen.returncode)
      log_method('Standard output:\n%s' % result_std)
      log_method('Error output:\n%s' % result_err)
    except:
      self.server.call('updatePartitionState', self.computer_id, 'reportError',
          'Unexpected issue while running buildout: %s:%s' % sys.exc_info()[:2] )
      raise
    if popen.returncode  != 0:
      self.server.call('updatePartitionState', self.computer_id, 'reportError',
          'Buildout finished with bad status code (%s). Stdout = "%s", stde'
          'rr = "%s"' % (popen.returncode, result_std, result_err))

  def updateInstanceProfiles(self):
    try:
      for instance in self.partition_dict_list:
        template_name = '%s-template.cfg' % instance['TYPE'].lower()\
          .replace(' ','-')
        template_path = self.findTemplate(template_name)
        if template_path is None:
          logging.warning('Template %r for %s not found, ignoring' % (
            template_name, instance['TYPE']))
          continue
        template_data = ''.join(file(template_path).readlines())

        template = PercentTemplate(template_data)
        if 'BT5_LIST' in instance:
          instance['BT5_LIST'] = '\n  '.join(instance['BT5_LIST'])
        if 'IP_ADDRESS_LIST' in instance:
          if instance['IP_ADDRESS_LIST']:
            instance['IP_ADDRESS'] = instance['IP_ADDRESS_LIST'][0]
        out = file(os.path.join(self.options.instances_directory, '%s.cfg' %
          instance['ID']), 'w')
        out.write(template.substitute(instance))
        out.close()
    except:
      self.server.call('updatePartitionState', self.computer_id, 'reportError',
          'Unexpected issue while updating instance profiles: %s:%s' %
          sys.exc_info()[:2])
      raise

  def findTemplate(self, template_name):
    for template_directory in self.template_directory_list:
      path = os.path.join(template_directory, template_name)
      if os.path.isfile(path):
        logging.debug('Found template %r' % path)
        return path
    return None

  def updateBaseProfile(self):
    # TODO:
    #  * cleanup in case of problem
    #  * use safe update of output file
    try:
      template_data = ''.join(file(self.findTemplate('main-template.cfg')).\
          readlines())
      template = PercentTemplate(template_data)
      replacement_dict = {
        'BASE_PROFILE': self.options.base_profile,
        'INSTANCE_PROFILE_LIST': [],
        'INSTANCE_PART_LIST': []
      }
      for instance in [q for q in self.partition_dict_list \
          if q['TYPE'] != 'Zope Instance']:
        profile_path = os.path.join(self.options.instances_directory,
            '%s.cfg' % instance['ID'])
        if not os.path.exists(profile_path):
          logging.warning('Profile %r not generated, ignoring' % profile_path)
          continue
        replacement_dict['INSTANCE_PROFILE_LIST'].append('%s/%s.cfg' % (
          self.options.instances_directory, instance['ID']))
        replacement_dict['INSTANCE_PART_LIST'].append(instance['ID'])

      for instance in [q for q in self.partition_dict_list \
          if q['TYPE'] == 'Zope Instance']:
        profile_path = os.path.join(self.options.instances_directory,
            '%s.cfg' % instance['ID'])
        if not os.path.exists(profile_path):
          logging.warning('Profile %r not generated' % profile_path)
          continue
        replacement_dict['INSTANCE_PROFILE_LIST'].append('%s/%s.cfg' % (
          self.options.instances_directory, instance['ID']))
        replacement_dict['INSTANCE_PART_LIST'].append(instance['ID'])

      replacement_dict['INSTANCE_PROFILE_LIST'] = '\n  '.join(
          replacement_dict['INSTANCE_PROFILE_LIST'])
      replacement_dict['INSTANCE_PART_LIST'] = '\n  '.join(
          replacement_dict['INSTANCE_PART_LIST'])
      out = file(self.options.main_output, 'w')
      out.write(template.substitute(replacement_dict))
      out.close()
    except:
      self.server.call('updatePartitionState', self.computer_id, 'reportError',
          'Unexpected issue while updating base profile: %s:%s' %
          sys.exc_info()[:2])
      raise

  def setRunning(self):
    if os.path.exists(self.options.pid_file):
      # Pid file is present
      logging.warning('Timmy already have the pid file %s' %
          self.options.pid_file)
      try:
        pid = int(open(self.options.pid_file, 'r').readline())
      except ValueError:
        pid = None
      # XXX This could use psutil library.
      if pid is not None and os.path.exists("/proc/%s" % pid):
        if 'timmy' in open('/proc/%s/cmdline' % pid, 'r').readline():
          # In case process is present, ignore.
          logging.critical('A Timmy process is running with pid %s' % pid)
          sys.exit(1)
      logging.info('Pid file %r was stale one, overwritten' %
          self.options.pid_file)
    # Start new process
    self.write_pid()

  def setFinished(self):
    os.remove(self.options.pid_file)

  def write_pid(self):
    pid = os.getpid()
    try:
      f = open(self.options.pid_file, 'w')
      f.write('%s' % pid)
      f.close()
    except (IOError, OSError):
      logging.critical('Timmy could not write pidfile %s' %
          self.options.pid_file)
      raise

  def parseOptions(self):
    parser = OptionParser()
    parser.add_option("-k", "--key-file",
      help="File with server key")

    parser.add_option("-s", "--server-url",
      help="URL of provisioning server")

    parser.add_option("-b", "--base-profile",
      help="Base profile")

    parser.add_option("-t", "--template-directory",
      help="Directory with templates")

    parser.add_option("-o", "--main-output",
      help="Main profile output, which will be run by buildout")

    parser.add_option("-d", "--instances-directory",
      help="Instances profiles output, must be existing directory")

    parser.add_option("-r", "--buildout-binary",
      help="Buildout binary to run")

    parser.add_option("-p", "--pid-file",
      help="Lock file with the pid file.")

    parser.add_option("-l", "--log-file",
      help="Log file.")

    parser.set_defaults(buildout_offline=False)
    parser.add_option("-N", "--buildout-offline", action="store_true",
      help="Run buildout in offline mode.")

    (options, args) = parser.parse_args()
    required_option_list = [
        'base_profile',
        'buildout_binary',
        'instances_directory',
        'key_file',
        'main_output',
        'server_url',
        'pid_file'
      ]
    missing_required_option_list = [o for o in required_option_list \
        if not getattr(options, o, None)]
    if len(missing_required_option_list):
      parser.error('Required options are missing: %s' % ', '.join(
        missing_required_option_list))

    for filepath in options.base_profile,:
      if not os.path.exists(filepath):
        raise ValueError('Cannot find "%s"' % os.path.abspath(filepath))
    if not os.path.exists(options.instances_directory):
      os.mkdir(options.instances_directory)
    elif not os.path.isdir(options.instances_directory):
      raise ValueError('File %s is not a directory' %
          os.path.abspath(options.instances_directory))
    return options, args

  def updateSupervisor(self):
    """Forces configuration reload of supervisor"""
    supervisorctl_popen = subprocess.Popen([SUPERVISORCTL, 'update'],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (result_std, result_err) = supervisorctl_popen.communicate()
    if supervisorctl_popen.returncode == 0:
      log_message = 'Supervisorctl updated with: stdout = %r stderr = %r' % (
          result_std, result_err)
      self.server.call('updatePartitionState', self.supervisor_id, 'reportStarted',
          log_message)
      logging.info(log_message)
    else:
      log_message = 'Supervisorctl issue during update: stdout = %r stderr ='\
          ' %r' % (result_std, result_err)
      self.server.call('updatePartitionState', self.supervisor_id, 'reportError',
          log_message)
      logging.error(log_message)

  def startAllSupervisored(self):
    """Invoke start on all supervisor controlled programs

    It is no-op in case of running program, but forces start of problematic
    services, for example in FATAL state"""
    supervisorctl_popen = subprocess.Popen([SUPERVISORCTL, 'start', 'all'],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (result_std, result_err) = supervisorctl_popen.communicate()
    if supervisorctl_popen.returncode != 0:
      log_message = 'Supervisorctl issue during start all: stdout = %r std'\
          'err = %r' % (result_std, result_err)
      self.server.call('updatePartitionState', self.supervisor_id, 'reportError',
          log_message)
      logging.error(log_message)
    elif len(result_std) > 0 or len(result_err) > 0:
      # emit log only in case if something important was done by supervisor
      # no need to pollute master server, as this is only local server hack
      logging.info('Supervisorctl start all with: stdout = %r stderr = %r' % (
          result_std, result_err))

def run():
  return Timmy().run()

if __name__ == '__main__':
  run()
